ROOT="/"
FIRMWARE_INSTALL_PATH="#{ROOT}/protonet/firmware/"
CONFIG_BASE="#{ROOT}/data/var/lib/protonet/config/"

require 'digest/md5'
require 'fileutils'
require 'open-uri'
require 'json'
require File.dirname(__FILE__) + '/protonet_config'

class FirmwareUpdater
  def initialize(version, url, kernel, image)
    @url = url
    @version = version
    @image = image
    @kernel = kernel
  end

  def firmware_base_url
    @url
  end

  def exists?(md5, path)
    (Digest::MD5.file(path).hexdigest == md5) rescue false
  end

  def download_and_check(md5, path)
    retry_count = 0
    puts  "downloading #{path}"
    FileUtils.mkdir_p(File.dirname(path))

    temp_file = "#{md5}.download_temp"
    loop do
      puts  "retrying #{path}" if retry_count >= 1

      # download tempfile based on md5
      system "curl -s -C - -o #{temp_file} #{firmware_base_url}/#{path} && mv #{temp_file} #{path}"

      break if retry_count >= 5 || exists?(md5, path)

      retry_count += 1
    end

    # throw exception if 3 retries are not enough
    raise unless exists?(md5, path)
  end


  def load_manifest
    image_files = open("#{firmware_base_url}/#{@version}/MANIFEST").
      read.split("\n").
      map { |line| line.split(/\s+/) }
  end

  def ensure_downloaded
    puts  "updating #{@version}"

    firmware_files = load_manifest

    firmware_files.each do |md5, path|
      next if exists?(md5, path)
      download_and_check(md5, path)
    end

    files_not_in_firmware = Dir["#{@version}/**/**"].reject { |path| File.directory?(path) }.sort - firmware_files.map { |i| i[1] }.sort

    if files_not_in_firmware.count > 0
      FileUtils.rm(files_not_in_firmware)      
      puts  "deleted #{files_not_in_firmware.join(", ")}"
    end


    true # all good!
  end

  def activate_kernel
    puts  "activating #{@kernel}"

    begin
      `dpkg --force-all -i #{@kernel}` 
      $?.exitstatus == 0
    rescue Exception => e
      puts "Cannot install kernel! #{e.inspect}"
      false
    end
  end

  def activate!
    Dir.chdir(FIRMWARE_INSTALL_PATH) do 
      puts "activating #{@version}"

      FileUtils.ln(@image, "#{@version}/image")
      FileUtils.rm_f('current')
      FileUtils.ln_s("./#{@version}", "current")
    end
    activate_kernel
  end
end

def generate_protonet_config
  config = protonet_config
  puts  "generating config"
  File.open("#{CONFIG_BASE}/current", "w") { |f| f.write JSON.pretty_generate(config) }
  FileUtils.chmod 0664, "#{CONFIG_BASE}/current"
  FileUtils.chown 'protonet', 'protonet', "#{CONFIG_BASE}/current"
  true
rescue Exception => e
  puts "ERROR: cannot generate config! #{e.inspect}"
  false
end


def prep_system
  FileUtils.mkdir_p(FIRMWARE_INSTALL_PATH)
  FileUtils.mkdir_p(CONFIG_BASE)

  Dir.chdir(ROOT) do 
    FileUtils.rm_f("firmware")
    FileUtils.ln_s("./protonet/firmware", "firmware")
  end

  true
end

def migrate_protonet_data
  return unless File.directory?('/home/protonet/dashboard')
  puts "migrate_protonet_data"

  `sv down /home/protonet/dashboard/shared/services/enabled/*`

  `mkdir -p /data/home/protonet/dashboard/shared`
  `mv /home/protonet/dashboard/shared/files /data/home/protonet/dashboard/shared`
  `mv /home/protonet/dashboard/shared/public /data/home/protonet/dashboard/shared`
  `mv /home/protonet/dashboard/shared/log /data/home/protonet/dashboard/shared`
  `mkdir -p /data/var/lib`
  `mv /var/lib/mysql /data/var/lib`
  `mv /var/lib/libvirt /data/var/lib`
  `mkdir -p /data/etc`
  `mv /etc/libvirt /data/etc`
  `mv /home/protonet/dashboard /home/protonet/inactive_dashboard`
  `mv /etc/sv/protonet_runit /etc/protonet_runit.inactive`
end

def undo_migrate_protonet_data
  return unless File.directory?('/home/protonet/inactive_dashboard')
  puts "undo_migrate_protonet_data"

  `mv /data/home/protonet/dashboard/shared/files /home/protonet/dashboard/shared/`
  `mv /data/home/protonet/dashboard/shared/public /home/protonet/dashboard/shared/`
  `mv /data/home/protonet/dashboard/shared/log /home/protonet/dashboard/shared/`
  `mv /data/var/lib/libvirt /var/lib`
  `mv /data/etc/libvirt /etc`
  `mv /home/protonet/inactive_dashboard /home/protonet/dashboard`
  `mv /etc/protonet_runit.inactive /etc/sv/protonet_runit`
end

def with_lockfile(lockfile)
  file = File.open(lockfile, 'w+')
  if file.flock(File::LOCK_NB | File::LOCK_EX)
    yield
  else
    puts "cannot aquire #{lockfile}"
  end
ensure
  file.close
end

def migrate_pre_firmware_system
  build = `cat /home/protonet/dashboard/current/BUILD`.match(/^(stable|beta)\/(\d+)$/)[2].to_i rescue nil
  
  if (build && build < 36)
    puts "migrating old firmware"
    generate_protonet_config && migrate_protonet_data
  else
    puts "no old firmware found, moving on to firmware activation"
    true
  end
end

prep_system

Dir.chdir(FIRMWARE_INSTALL_PATH) do 
  with_lockfile('firmware_updater.lock') do
    updater = FirmwareUpdater.new(*ARGV)
    updater.ensure_downloaded && migrate_pre_firmware_system && updater.activate! && exit(0)
  end
end

exit(1)
